import aiohttp
import asyncio
import async_timeout
import json
import logging
import ssl

import diax.client
import diax.errors

LOGGER = logging.getLogger(__name__)

class AsyncClient(diax.client.Client):
    def __init__(self, *args, certificate_bundle=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.certificate_bundle = certificate_bundle
        # We can't create the session here because we'll get a lot message
        # about it not being inside a coroutine, even though it's synchronous
        self.session = None

    async def streamevents(self):
        url = self.url('events', '/')
        buffer_ = b''
        while True:
            try:
                LOGGER.info("Connecting to %s", url)
                async with self.session.get(url) as response:
                    while True:
                        buffer_ += await response.content.read(256)
                        try:
                            text = buffer_.decode('utf-8')
                        except ValueError:
                            LOGGER.debug("Failed to decode event buffer as UTF-8, waiting for more")
                            continue
                        if '\n' not in text:
                            LOGGER.debug("Failed to find \\n in %s, waiting for more", buffer_)
                            continue
                        datagram = text[:text.index('\n') + 1]
                        datagram_bytes = datagram.encode('utf-8')
                        LOGGER.debug("Found %s bytes of message data in %s bytes buffer", len(datagram_bytes), len(buffer_))
                        buffer_ = buffer_[len(datagram_bytes):]
                        await json.loads(datagram)
            except asyncio.TimeoutError as e:
                LOGGER.warning("Error streaming events: %s", e)

    async def _ensure_session(self):
        if not self.session:
            if self.certificate_bundle:
                context = ssl.create_default_context(
                    cafile  = self.certificate_bundle
                )
                connector = aiohttp.TCPConnector(ssl_context=context)
            else:
                connector = aiohttp.TCPConnector()
            kwargs = {'connector': connector}
            if self.internal_token:
                kwargs['auth']  = aiohttp.BasicAuth('api', self.internal_token)
            self.session = aiohttp.ClientSession(**kwargs)

    async def login(self):
        await self._ensure_session()
        self._ensure_has_credentials()
        await self._do_login()
        await self._maybe_impersonate()

    async def _do_login(self):
        if self.internal_token:
            # do something with aiohttp.ClientSession(headers={'Authorization': "Basic ..."})
            raise Exception("Sorry, I haven't implemented internal tokens in aioclient")
        payload = {
            'password'  : self.password,
            'username'  : self.username,
        }
        await self.post('users', '/sessions/', payload)

    async def _maybe_impersonate(self):
        if self.impersonate:
            raise Exception("Sorry, I haven't implemented impersonation yet")

    async def post(self, service, path, payload):
        url = self.url(service, path)
        with async_timeout.timeout(10):
            async with self.session.post(url, data=json.dumps(payload), headers={'Content-Type': 'application/json'}) as response:
                return await self._validate_response('POST', url, response)

    async def rawput(self, url, data, headers):
        async with self.session.put(url, data=data, headers=headers) as response:
            return await self._validate_response('PUT', url, response)

    async def rawget(self, url):
        await self._ensure_session()
        async with self.session.get(url) as response:
            return await self._validate_response('GET', url, response)

    async def _validate_response(self, method, url, response):
        text = await response.text()
        if not (response.status >= 200 and response.status < 300):
            raise diax.errors.RequestError(method, url, response)
        try:
            return json.loads(text), response.headers
        except ValueError:
            return text, response.headers

def create(args):
    "Create a client to talk to 3DIAX with the specified environment"
    return AsyncClient(
        certificate_bundle  = args.certificate_bundle,
        environment         = args.environment,
        impersonate         = args.impersonate,
        internal_token      = args.internal_token,
        password            = args.password,
        username            = args.username,
    )




import argparse
import logging
import pprint

import diax.client
import diax.errors
import diax.scripting
import diax.services

LOGGER = logging.getLogger(__name__)
def _create(client, service, args):
    payload = _payload_from_args(args)
    LOGGER.info("Creating new %s resource in service %s with parameters %s", args.service, args.resource, pprint.pformat(payload))
    try:
        response = service[args.resource].post(payload)
        LOGGER.info("Created %s %s", args.resource, response)
        return 0
    except diax.errors.ValidationError as e:
        LOGGER.error("Can't create %s: %s", args.resource, e)
        return 1

def _delete(client, service, args):
    LOGGER.info("Deleting %s", args.resource)
    results = client.rawdelete(args.resource)
    LOGGER.info("Done")

def _list(client, service, args):
    LOGGER.info("Listing %s resources from %s", args.resource, args.service)
    results = service[args.resource].list()
    LOGGER.info("Results: %s", pprint.pformat(results))

def _payload_from_args(args):
    payload = {}
    for arg in args.payload:
        if '=' not in arg:
            raise Exception("You supplied the value '{arg}'. Did you mean '{arg}=something'?".format(arg=arg))
        k, _, v = arg.partition('=')
        payload[k] = v
        if payload[k] == 'true':
            payload[k] = True
        if payload[k] == 'false':
            payload[k] = False
        if payload[k] == 'null':
            payload[k] = None
    return payload

def _update(client, service, args):
    payload = _payload_from_args(args)
    LOGGER.info("Updating %s with payload %s", args.resource, pprint.pformat(payload))
    results = client.rawput(args.resource, payload)
    LOGGER.info("Done")

def main():
    logging.basicConfig(level=logging.DEBUG)
    parser = diax.scripting.parser()
    subparsers = parser.add_subparsers(help='The command to perform')
    subparsers.required = True
    subparsers.dest = 'command'

    parser_create = subparsers.add_parser('create', help="Create a new resource")
    parser_create.set_defaults(command=_create)
    parser_create.add_argument('service',
        choices=['data', 'erp', 'quickslice', 'tetra', 'woodhouse', 'users'],
        help="The name of the service to interact with",
    )
    parser_create.add_argument('resource',
        help="The path of the resource to create such as '/model/'",
    )
    parser_create.add_argument(
        'payload',
        nargs=argparse.REMAINDER,
        help="The values to send in creating the resource with the form foo=bar",
    )

    parser_list = subparsers.add_parser('list', help="List all available resources of the given type")
    parser_list.set_defaults(command=_list)
    parser_list.add_argument('service',
        choices=['data', 'erp', 'quickslice', 'tetra', 'woodhouse', 'users'],
        help="The name of the service to interact with",
    )
    parser_list.add_argument('resource',
        help="The path of the resource to list such as '/model/'",
    )

    parser_delete = subparsers.add_parser('delete', help='Delete something')
    parser_delete.add_argument('resource', help='The URI of the resource to be deleted')
    parser_delete.set_defaults(command=_delete)

    parser_update = subparsers.add_parser('update', help='Update a resource')
    parser_update.set_defaults(command=_update)
    parser_update.add_argument('resource', help='The URI of the resource to be updated')
    parser_update.add_argument(
        'payload',
        nargs=argparse.REMAINDER,
        help="The values to send in updating the resource with the form foo=bar",
    )
    args = parser.parse_args()

    client = diax.client.create(args)
    client.login()

    if hasattr(args, 'service'):
        service = diax.services.connect(client, args.service)
    else:
        service = None


    return args.command(client, service, args)



import diax.errors

def get(client, user):
    "Get information on the user provided an email, a URI, etc"
    if user.startswith('http'):
        data, _ = client.rawget(user)
        return data
    result = client.list('users', '/users/', {'username': user})
    if len(result) == 0:
        result = client.list('users', '/users/', {'email': user})
        if len(result) == 0:
            raise diax.errors.UserNotFound("Could not find a user with username or email '{}'".format(user))
        return result[0]
    elif len(result) == 1:
        return result[0]
    else:
        raise Exception("Didn't expect to get more than one user with username '{}'".format(user))



import urllib.parse
import uuid

def _maybe_uuid(value):
    try:
        return uuid.UUID(value)
    except:
        pass

def is_uri(value, service=None, path=None):
    url = urllib.parse.urlparse(value)
    if url.scheme != 'https':
        return False
    if service and not service in url.netloc:
        return False
    if path and not url.path.startswith(path):
        return False
    return True

def extract_uuids(value):
    url = urllib.parse.urlparse(value)
    parts = url.path.split('/')
    uuids = [_maybe_uuid(value) for value in parts]
    return [x for x in uuids if x]



import logging
import os

LOGGER = logging.getLogger(__name__)

class UploadError(Exception):
    "Failure to upload data"

def upload(client, filename, type_='stl', name=None):
    name = name or os.path.basename(filename)
    payload = {
        'name'  : name,
        'type'  : type_,
    }
    _, headers = client.post('models', '/model/', payload)
    model = headers['Location']
    upload_url = headers['X-Upload-Location']
    LOGGER.info("Created model %s, uploading data from %s", model, filename)
    with open(filename, 'rb') as f:
        response = client.session.put(upload_url, data=f, headers={'Content-Type': 'application/octet-stream'})
        if not response.ok:
            raise UploadError("Failed to upload model data from {} to {}: {}".format(filename, upload_url, response.text))

    LOGGER.debug("Upload to %s complete", upload_url)
    return model



import json
import logging
import os
import pprint
import requests
import urllib
import yaml

import diax.errors
import diax.uri

LOGGER = logging.getLogger(__name__)

CREDENTIALS_FILE = os.path.join(os.environ['HOME'], '.authentise', 'credentials')
class Client():
    def __init__(self, environment, impersonate=None, internal_token=None, password=None, ssl=True, ssl_verify=True, username=None,):
        self.environment    = environment
        self.impersonate    = impersonate
        self.internal_token = internal_token
        self.password       = password
        self.session        = requests.Session()
        self.ssl            = ssl
        self.ssl_verify     = ssl_verify
        self.username       = username

    def _do_login(self):
        if self.internal_token:
            LOGGER.info("No login necessary, using internal token")
            self.session.auth = ('api', self.internal_token)
            return
        payload = {
            'password'  : self.password,
            'username'  : self.username,
        }
        _, headers = self.post('users', '/sessions/', payload)
        LOGGER.debug("Got headers from login: %s", pprint.pformat(headers.items()))

    def _auth(self):
        if self.internal_token:
            return ('api', self.internal_token)
        else:
            return None

    def _ensure_has_credentials(self):
        if self.internal_token:
            LOGGER.info("Using internal token")
            return
        if self.username and self.password:
            LOGGER.debug("Using credentials provided via commandline")
            return
        if os.path.exists(CREDENTIALS_FILE):
            with open(CREDENTIALS_FILE, 'r') as f:
                data = yaml.load(f)
            credentials = data.get(self.environment)
            if not credentials:
                raise diax.errors.CredentialsError("You have a {} file but it needs a section for environment {}".format(CREDENTIALS_FILE, self.environment))
            self.password = credentials.get('password')
            self.username = credentials.get('username')
            if not self.username:
                raise diax.errors.CredentialsError("You have a section in {} for environment {} but it doesn't have a username".format(CREDENTIALS_FILE, self.environment))
            if not self.password:
                raise diax.errors.CredentialsError("You have a section in {} for environment {} but it doesn't have a password".format(CREDENTIALS_FILE, self.environment))
            return
        raise diax.errors.CredentialsError("You need to either specify credentials on the commandline with -u and -p or you need to create a credentials file at {}".format(CREDENTIALS_FILE))

    def _get_user_uri(self, username):
        results = self.list('users', '/users/', filters={'username': username})
        if results:
            return results[0]['uri']

    def _maybe_impersonate(self):
        if not self.impersonate:
            return
        LOGGER.debug("Impersonating %s", self.impersonate)
        uri = self.impersonate if diax.uri.is_uri(self.impersonate, 'data', '/users/') else self._get_user_uri(self.impersonate)
        if not uri:
            raise diax.errors.ImpersonationError("The value '{}' is not a user URI or username".format(self.impersonate))
        payload = {
            'impersonating' : uri,
        }
        self.put('users', '/sessions/0/', payload)

    def _validate(self, response, method, url):
        if not response.ok:
            LOGGER.debug("Response body: %s", response.text)
            raise diax.errors.RequestError(method, url, response)
        if response.status_code == 204:
            return None, response.headers
        try:
            return response.json(), response.headers
        except ValueError:
            return response.text, response.headers

    def get(self, service, path, filters={}):
        url = self.url(service, path, filters)
        response = self.session.get(url, verify=self.ssl_verify)
        out, _ = self._validate(response, 'GET', url)
        return out

    def list(self, service, path, filters={}):
        result = self.get(service, path, filters=filters)
        return result['resources']

    def login(self):
        self._ensure_has_credentials()
        self._do_login()
        self._maybe_impersonate()

    def options(self, service, path):
        url = self.url(service, path)
        response = self.session.options(url, verify=self.ssl_verify)
        out, _ = self._validate(response, 'OPTIONS', url)
        return out

    def post(self, service, path, payload):
        url = self.url(service, path)
        response = self.session.post(url, json=payload, verify=self.ssl_verify, auth=self._auth())
        out, headers = self._validate(response, 'POST', url)
        return out, headers

    def put(self, service, path, payload):
        url = self.url(service, path)
        response = self.session.put(url, json=payload, verify=self.ssl_verify)
        out, _ = self._validate(response, 'PUT', url)
        return out

    def rawdelete(self, url):
        response = self.session.delete(url, verify=self.ssl_verify)
        out, _ = self._validate(response, 'DELETE', url)
        return out

    def rawpost(self, url, payload, headers=None):
        response = self.session.post(url, json=payload, verify=self.ssl_verify, headers=headers)
        _, headers = self._validate(response, 'POST', url)
        return headers

    def rawput(self, url, payload=None, json=None, headers=None):
        response = self.session.put(url, payload, json=json, verify=self.ssl_verify, headers=headers)
        out, _ = self._validate(response, 'PUT', url)
        return out

    def rawget(self, url):
        return self._validate(self.session.get(url, verify=self.ssl_verify, auth=self._auth()), 'GET', url)

    def url(self, service, path, filters=None):
        base_pattern = "{{}}.{}".format(self.environment) if "{}" not in self.environment else self.environment
        base = base_pattern.format(service)
        _url = "{}://{}{}".format('https' if self.ssl else 'http', base, path)
        queryargs = {}
        filters = filters or {}
        queryargs.update({'filter[{}]'.format(k): v for k, v in filters.items()})
        if queryargs:
            _url += '?' + urllib.parse.urlencode(queryargs)
        return _url

    def streamevents(self):
        _url = self.url('events', '/')
        response = self.session.get(_url, stream=True, verify=self.ssl_verify)
        if not response.ok:
            raise diax.errors.RequestError('GET', _url, response)
        for line in response.iter_lines():
            if line:
                line = line.decode('utf-8')
                yield json.loads(line)

def create(args):
    "Create a client to talk to 3DIAX with the specified environment"
    return Client(
        environment         = args.environment,
        impersonate         = args.impersonate,
        internal_token      = args.internal_token,
        ssl                 = args.ssl,
        ssl_verify          = args.certificate_bundle or not args.ssl_no_verify,
        password            = args.password,
        username            = args.username,
    )




class DiaxError(Exception):
    "Root exception for anything in diax"

class CredentialsError(DiaxError):
    "Something went wrong with your credentials, likely they weren't provided"

class RequestError(DiaxError):
    "A request to Diax failed in some way"
    def __init__(self, method, url, response):
        super().__init__()
        self.method   = method
        self.response = response
        self.url      = url

    def __str__(self):
        return "RequestError on {} to {} returned {}: {}".format(
            self.method,
            self.url,
            self.response.status_code,
            self.response.text,
        )


class ImpersonationError(DiaxError):
    "Impersonation failed for some reason"

class ValidationError(DiaxError):
    "The request failed validation"

class UserNotFound(DiaxError):
    "The specified users does not exist"



#!/usr/bin/env python
import contextlib
import imp
import os
import re
import subprocess

from setuptools import setup

DATA_ROOTS = ['alembic']
PROJECT = 'diax'
VERSION_FILE = 'diax/__init__.py'

def _get_output_or_none(args):
    try:
        return subprocess.check_output(args).decode('utf-8').strip()
    except subprocess.CalledProcessError:
        return None

def _get_git_description():
    return _get_output_or_none(['git', 'describe'])

def _get_git_branches_for_this_commit():
    branches = _get_output_or_none(['git', 'branch', '-r', '--contains', 'HEAD'])
    split = branches.split('\n') if branches else []
    return [branch.strip() for branch in split]

def _is_on_releasable_branch(branches):
    return any([branch == 'origin/master' or branch.startswith('origin/hotfix') for branch in branches])

def _git_to_version(git):
    match = re.match(r'(?P<tag>[\d\.]+)-(?P<offset>[\d]+)-(?P<sha>\w{8})', git)
    if not match:
        version = git
    else:
        version = "{tag}.post0.dev{offset}".format(**match.groupdict())
    print("Calculated {} version '{}' from git description '{}'".format(PROJECT, version, git))
    return version

@contextlib.contextmanager
def write_version():
    git_description = _get_git_description()
    git_branches = _get_git_branches_for_this_commit()
    version = _git_to_version(git_description) if git_description else None
    if git_branches and not _is_on_releasable_branch(git_branches):
        print("Forcing version to 0.0.1 because this commit is on branches {} and not a whitelisted branch".format(git_branches))
        version = '0.0.1'
    if version:
        with open(VERSION_FILE, 'r') as version_file:
            old_contents = version_file.read()
        with open(VERSION_FILE, 'w') as version_file:
            version_file.write('__version__ = "{}"\n'.format(version))
    yield
    if version:
        with open(VERSION_FILE, 'w') as version_file:
            version_file.write(old_contents)

def get_version():
    basedir = os.path.abspath(os.path.dirname(__file__))
    version = imp.load_source('diax', os.path.join(basedir, PROJECT, '__init__.py'))
    return version.__version__

def get_data_files():
    data_files = []
    for data_root in DATA_ROOTS:
        for root, _, files in os.walk(data_root):
            data_files.append((os.path.join(PROJECT, root), [os.path.join(root, f) for f in files]))
    return data_files

def main():
    with write_version():
        setup(
            name='diax',
            version=get_version(),
            description="The official client to the 3DIAX API",
            long_description=open('README.md').read(),
            author='Authentise, Inc.',
            author_email='engineering@authentise.com',
            install_requires=[
                'ramlfications==0.1.9',
                'requests==2.12.4',
            ],
            extras_require={
                'develop': [
                    'astroid==1.4.4',
                    'coverage==3.7.1',
                    'mothermayi>=0.4',
                    'mothermayi-isort>=0.8',
                    'mothermayi-pylint>=0.5',
                    'py==1.4.31',
                    'pylint==1.5.4',
                    'pytest==3.0.2',
                    'pytest-cov==2.3.1',
                    'pytest-raises>=0.6'
                ],
            },
            packages=[
                'diax',
            ],
            package_data={
                'diax'                : ['diax/*'],
            },
            data_files          = get_data_files(),
            scripts             = ['bin/3diax'],
            include_package_data=True,
        )

if __name__ == '__main__':
    main()



import argparse
import logging
import sys

import diax.log

def excepthook(_type, value, tb):
    import pdb
    import traceback
    traceback.print_exception(_type, value, tb)
    pdb.pm()

class ArgumentParserDiax(argparse.ArgumentParser):
    def parse_args(self):
        results = super().parse_args()
        if results.pdb:
            sys.excepthook = excepthook
        diax.log.setup(level=logging.DEBUG if results.verbose else logging.INFO)
        return results


def parser(*args, **kwargs):
    _parser = ArgumentParserDiax(*args, **kwargs)
    _parser.add_argument('--impersonate',
        dest='impersonate',
        help=("Impersonate another user in 3DIAX. This will only work if your user credentials "
            "are allowed to impersonate, otherwise login will fail entirely"),
    )
    _parser.add_argument('-e', '--env', '--environment',
        dest='environment',
        default='authentise.com',
        help=("The base URL of the environment to interact with. This should be something like "
            "'dev-auth.com' or 'authentise.com'. Service URLs will be built from this base"),
    )
    _parser.add_argument('-u', '--username',
        dest='username',
        help=("The username to use when logging in to 3DIAX"),
    )
    _parser.add_argument('-p', '--password',
        dest='password',
        help=("The password to use when logging in to 3DIAX"),
    )
    _parser.add_argument('--internal-token',
        dest='internal_token',
        help=("The internal token for the installation which grants superuser access"),
    )
    _parser.add_argument('--no-ssl',
        action='store_false',
        dest='ssl',
        help=("When provided disable using SSL for all communications"),
    )
    _parser.add_argument('--pdb',
        action='store_true',
        dest='pdb',
        help=("When provided, drop into a debugger on unhandled exception"),
    )
    _parser.add_argument('-v',
        action='store_true',
        dest='verbose',
        help=("When provided show more verbose logging"),
    )
    _parser.add_argument('--ssl-no-verify',
        action='store_true',
        dest='ssl_no_verify',
        help=("When provided do not verify SSL certificates. Useful for local dev work"),
    )
    _parser.add_argument('--certificate-bundle',
        default=None,
        dest='certificate_bundle',
        help=("When provided, use this certificate bundle to verify sSL"),
    )
    return _parser



import json
import logging

import diax.errors
import ramlfications
import requests

LOGGER = logging.getLogger(__name__)

def _to_resource_name(uri):
    "Convert a resource uri like '/foo/bar/' to a name like foo.bar"
    if uri[0] == '/':
        uri = uri[1:]
    if uri[-1] == '/':
        uri = uri[:-1]
    parts = uri.split('/')
    valid = [p for p in parts if p != '{uuid}']
    return '.'.join(valid)

class Service():
    def __init__(self, client, name):
        self.baseUri        = None
        self.client         = client
        self.documentation  = None
        self.media_type     = None
        self.name           = name
        self.options        = None
        self.resources      = {}
        self.title          = None

    def load(self):
        try:
            response = self.client.options(self.name, '/')
        except diax.errors.RequestError as e:
            LOGGER.error("Failed to load up RAML for %s: %s", self.name, e)
            return
        raml = ramlfications.loads(response)
        for k in ('baseUri', 'documentation', 'mediaType', 'title', 'version'):
            setattr(self, k, raml.pop(k, None))
        for relativeUri, resource in raml.items():
            resource_name = _to_resource_name(relativeUri)
            if resource_name not in self.resources:
                self.resources[resource_name] = Resource(self, relativeUri)
            self.resources[resource_name].add(resource)

    def __getitem__(self, name):
        try:
            return self.resources[name]
        except KeyError:
            LOGGER.warning("Unable to find resource %s in service %s. Attempting to fake one...", name, self.name)
            return Resource(self, '/{}/'.format(name))

class Resource():
    def __init__(self, service, uri):
        self.methods = {}
        self.service = service
        self.uri     = uri

    def add(self, schema):
        for k, v in schema.items():
            if k not in ('delete', 'get', 'options', 'patch', 'post', 'put'):
                continue
            self.methods[k] = Method(v)

    def post(self, payload):
        try:
            self.methods['post'].validate(payload)
        except KeyError:
            LOGGER.warning("Can't validate paramters")
        _, headers = self.service.client.post(self.service.name, self.uri, payload)
        return headers['Location']

    def list(self):
        data = self.service.client.get(self.service.name, self.uri)
        return data

class Method():
    def __init__(self, schema):
        self.schema = None
        for content_type, descriptor in schema.get('body', {}).items():
            if not (content_type == 'application/json' and 'schema' in descriptor):
                continue
            self.schema = json.loads(descriptor['schema'])

    def validate(self, payload):
        errors = []
        for k in self.schema['required']:
            if k not in payload:
                errors.append("The parameter '{}' is required".format(k))
        if errors:
            raise diax.errors.ValidationError("\n".join(errors))

def connect(client, name):
    "Connect to the given service, pull down the definition of its endpoints"
    service = Service(client, name)
    service.load()
    return service
